#include <unordered_map>

#include <maya/MPxLocatorNode.h>
#include <maya/MIntArray.h>
#include <maya/MUintArray.h>
#include <maya/MFloatArray.h>
#include <maya/MDoubleArray.h>
#include <maya/MFloatPointArray.h>
#include <maya/MFloatVectorArray.h>

#include <maya/MUserData.h>
#include <maya/MDrawRegistry.h>
#include <maya/MPxDrawOverride.h>
#include <maya/MDrawContext.h>
#include <maya/MHWGeometryUtilities.h>

class gpuCacher : public MPxLocatorNode
{
public:
	gpuCacher();
	virtual ~gpuCacher();

	virtual MStatus compute(const MPlug& plug, MDataBlock& data);
	virtual void draw(M3dView &view, const MDagPath &path, M3dView::DisplayStyle style, M3dView::DisplayStatus status);
	virtual void postConstructor();

	static void * creator();
	static MStatus initialize();

	static MTypeId id;
	static MString drawDbClassification;
	static MString drawRegistrantId;
	static MObject aInBaseAttr;
	static MObject aInData;
		static MObject aInAttr;
		static MObject aInMesh;
		static MObject aColor;
	static MObject aExclusiveCamera;
	static MObject aInTime;
	static MObject aSceneStateTracking;
	static MObject aReset;
	static MObject aCull;
	static MObject aSpecClrR;
	static MObject aSpecClrG;
	static MObject aSpecClrB;
	static MObject aIsFrameCached;

public:
	MObject _oThisNode;

	bool _glew;
	GLuint _vboGeo, _vboIds;

	double _t;

	bool _ifc, _cull;
	std::unordered_map<unsigned int, std::unordered_map<double, bool> > _mapIfc;

	std::unordered_map<double, MFloatArray> _mapBaseAttr;
	std::unordered_map<unsigned int, std::unordered_map<double, MFloatArray> > _mapAttr;

	std::unordered_map<double, MFloatArray> _mapGeo;
	std::unordered_map<double, MUintArray> _mapIds;
	std::unordered_map<double, MUintArray> _mapRng;

	unsigned int _nGeoCnt, _nIdsCnt, _nCachedGeoCnt, _nCachedIdsCnt;
	GLfloat *_geo;
	GLuint *_ids;

	void vbo_create();
	void vbo_init();
	void vbo_update();
	void vbo_delete();

	void cacheBaseAttr(MArrayDataHandle& adh);
	void cacheAttr();
	void cacheGeo();
	void reset();

	float _scr, _scb, _scg;
};



class GpuCacherData : public MUserData
{
public:
	GpuCacherData() : MUserData(true) {}
	virtual ~GpuCacherData() {}

	MDagPath dpObj, dpCam;
};

class GpuCacherDrawOverride : public MHWRender::MPxDrawOverride
{
public:
	static MHWRender::MPxDrawOverride* Creator(const MObject& obj)
	{
		return new GpuCacherDrawOverride(obj);
	}

	virtual ~GpuCacherDrawOverride();

	virtual MHWRender::DrawAPI supportedDrawAPIs() const;

	virtual MUserData* prepareForDraw(
		const MDagPath& objPath,
		const MDagPath& cameraPath,
		const MHWRender::MFrameContext& frameContext,
		MUserData* oldData);

	virtual bool hasUIDrawables() const {return true;}

	virtual void addUIDrawables(
		const MDagPath& objPath,
		MHWRender::MUIDrawManager& drawManager,
		const MHWRender::MFrameContext& frameContext,
		const MUserData* data);

	static void draw(const MHWRender::MDrawContext& context, const MUserData* data);

private:
	GpuCacherDrawOverride(const MObject& obj);
};
